import numpy as np


class ParameterInitialization:

    """
    This Class initializes random values for initial weights and biases of our neural network
    """

    def __init__(self, input_size, n_layers, n_neurons_per_layer, n_classes):

        """
        input_size is the number of features or columns
        hidden_layers is the number of hidden layers of the network
        n_neurons is the number of neurons for the hidden layers
        n_classes is the number of class labels
        """
        
        self.input_size = input_size
        self.n_layers = n_layers
        self.n_neurons_per_layer = n_neurons_per_layer
        self.n_classes = n_classes

    def initialize_parameters(self):

        weights = {}
        biases = {}

        for i in range(self.n_layers):

            if i == 0:

                weights[f"W{i + 1}"] = np.random.rand(self.n_neurons_per_layer[i], self.input_size) - 0.5
                biases[f"b{i + 1}"] = np.random.rand(self.n_neurons_per_layer[i], 1) - 0.5

            else:
                
                weights[f"W{i + 1}"] = np.random.rand(self.n_neurons_per_layer[i], self.n_neurons_per_layer[i-1]) - 0.5
                biases[f"b{i + 1}"] = np.random.rand(self.n_neurons_per_layer[i], 1) - 0.5

        weights[f"W{self.n_layers + 1}"] = np.random.rand(self.n_classes, self.n_neurons_per_layer[-1]) - 0.5
        biases[f"b{self.n_layers + 1}"] = np.random.rand(self.n_classes, 1) - 0.5

        return weights, biases



class UpdateParameters:

    """
    This class defines the cost function to update weights and biases of our neural network
    """

    def update_parameters(self, weights, biases, s, alpha):

        p1 = {}
        p2 = {}

        l = len(weights)

        for i in range(1, l + 1):

            p1[f"W{i}"] = weights[f"W{i}"] - alpha * s[f"dW{i}"]
            p2[f"b{i}"] = biases[f"b{i}"] - alpha * s[f"db{i}"]

        return p1, p2


from NeuroInsight.activation_function import ActivationFunction
activation_function = ActivationFunction()


class ForwardPropagationForClassification:
    
    """
    This class is responsible for passing our data points through the neural network
    It uses ReLU for hidden layers and softmax for final output layer as activation function
    """

    def forward_propagation_classification(self, weights, biases, X):

        l = len(weights)
        p = {}

        for i in range(1, l+1):

            if i == 1 :

                p[f"Z{i}"] = weights['W1'].dot(X) + biases['b1']
                p[f"A{i}"] = activation_function.ReLU( weights['W1'].dot(X) + biases['b1'])

            elif i > 1 and i < l:

                p[f"Z{i}"] = weights[f"W{i}"].dot(p[f"A{i-1}"]) + biases[f"b{i}"]
                p[f"A{i}"] = activation_function.ReLU(weights[f"W{i}"].dot(p[f"A{i-1}"]) + biases[f"b{i}"])

            elif i == l:
                
                p[f"Z{i}"] = weights[f"W{i}"].dot(p[f"A{i-1}"]) + biases[f"b{i}"]
                p[f"A{i}"] = activation_function.Softmax(weights[f"W{i}"].dot(p[f"A{i-1}"]) + biases[f"b{i}"])

        return p
    


class ForwardPropagationForRegression:
    
    """
    This class is responsible for passing our data points through the neural network
    It uses ReLU for hidden layers and Linear for final output layer as activation function
    """

    def forward_propagation_regression(self, weights, biases, X):

        l = len(weights)
        p = {}

        for i in range(1, l+1):

            if i == 1 :

                p[f"Z{i}"] = weights['W1'].dot(X) + biases['b1']
                p[f"A{i}"] = activation_function.ReLU( weights['W1'].dot(X) + biases['b1'])

            elif i > 1 and i < l:

                p[f"Z{i}"] = weights[f"W{i}"].dot(p[f"A{i-1}"]) + biases[f"b{i}"]
                p[f"A{i}"] = activation_function.ReLU(weights[f"W{i}"].dot(p[f"A{i-1}"]) + biases[f"b{i}"])

            elif i == l:
                
                p[f"Z{i}"] = weights[f"W{i}"].dot(p[f"A{i-1}"]) + biases[f"b{i}"]
                p[f"A{i}"] = activation_function.Linear(p[f"Z{i}"])

        return p


import numpy as np

from NeuroInsight.activation_function import ActivationFunction
activation_function = ActivationFunction()

from NeuroInsight.one_hot_encoding import OneHotEncoding
onehotencoding = OneHotEncoding()

class BackPropagationForClassification:
    

    def back_propagation_classification(self, a, weights, X_train, Y_train, m):

        one_hot_Y = onehotencoding.one_hot_encoding(Y_train)
        p = {}
        l = len(weights)

        for i in range(l, -1+1, -1):

            if i == l:

                p[f"dZ{i}"] = a[f"A{i}"] - one_hot_Y
                p[f"dW{i}"] = 1 / m * p[f"dZ{i}"].dot(a[f"A{i-1}"].T)
                p[f"db{i}"] = 1 / m * np.sum(p[f"dZ{i}"], axis=1, keepdims=True)

            elif i < l and i > 1:

                p[f"dZ{i}"] = weights[f"W{i+1}"].T.dot(p[f"dZ{i+1}"]) * activation_function.derivative_ReLU(a[f"Z{i}"])
                p[f"dW{i}"] = 1 / m * p[f"dZ{i}"].dot(a[f"A{i-1}"].T)
                p[f"db{i}"] = 1 / m * np.sum(p[f"dZ{i}"], axis=1, keepdims=True)

            elif i == 1:

                p[f"dZ{i}"] = weights[f"W{i+1}"].T.dot(p[f"dZ{i+1}"]) * activation_function.derivative_ReLU(a[f"Z{i}"])
                p[f"dW{i}"] = 1 / m * p[f"dZ{i}"].dot(X_train.T)
                p[f"db{i}"] = 1 / m * np.sum(p[f"dZ{i}"], axis=1, keepdims=True)

        return  p
    


class BackPropagationForRegression:


    def back_propagation_regression(self, a, weights, X_train, Y_train, m):

        p = {}
        l = len(weights)

        for i in range(l, -1+1, -1):

            if i == l:

                p[f"dZ{i}"] = a[f"A{i}"] - Y_train
                p[f"dW{i}"] = 1 / m * p[f"dZ{i}"].dot(a[f"A{i-1}"].T)
                p[f"db{i}"] = 1 / m * np.sum(p[f"dZ{i}"])

            elif i < l and i > 1:

                p[f"dZ{i}"] = weights[f"W{i+1}"].T.dot(p[f"dZ{i+1}"]) * activation_function.derivative_ReLU(a[f"Z{i}"])
                p[f"dW{i}"] = 1 / m * p[f"dZ{i}"].dot(a[f"A{i-1}"].T)
                p[f"db{i}"] = 1 / m * np.sum(p[f"dZ{i}"])

            elif i == 1:
                
                p[f"dZ{i}"] = weights[f"W{i+1}"].T.dot(p[f"dZ{i+1}"]) * activation_function.derivative_ReLU(a[f"Z{i}"])
                p[f"dW{i}"] = 1 / m * p[f"dZ{i}"].dot(X_train.T)
                p[f"db{i}"] = 1 / m * np.sum(p[f"dZ{i}"])

        return  p



import numpy as np

class ActivationFunction:
    """
    This class defines the activation functions used in the neural network
    """

    def ReLU(self, Z):

        """
        For each element in Z, the ReLU function returns Z if Z is positive and return 0 if Z is negative
        """

        return np.maximum(Z ,0)
    
    
    def Softmax(self, Z):

        """
        The Softmax function converts the final score into probability distribution over different classes
        np.exp(Z) calculates the element-wise exponential of input Z
        np.sum(np.exp(Z)) calculates the sum of the exponentiated values of ell elements of Z
        """

        A = np.exp(Z) / sum(np.exp(Z))
        return A


    def Linear(self, Z):
        """
        Linear activation function for regression problem
        """
        return Z
    

    def derivative_ReLU(self, Z):

        """
        This derivative is used to compute the gradients of the loss with respect to the input of the ReLU activation function to optimize our parameters
        If any element in Z is greater than 0 , return 1 otherwise 0
        """

        return Z > 0


    



import numpy as np

from NeuroInsight.forward_propagation import ForwardPropagationForClassification
forward_propagation_c = ForwardPropagationForClassification()

from NeuroInsight.forward_propagation import ForwardPropagationForRegression
forward_propagation_r = ForwardPropagationForRegression()


class Evaluation:

    """
    This class defines some functions to evaluate the trained model
    """


    def predictions(self, A):
        return np.argmax(A, 0)
    

    def accuracy(self, predictions, Y):
        return np.sum(predictions == Y) / Y.size
    
    
    def classification_validation_predictions(self, X, weights, biases):
        forward_output = forward_propagation_c.forward_propagation_classification(weights, biases, X)
        key, value = list(forward_output.items())[-1]
        prediction = self.predictions(value)
        return prediction
    

    def regression_validation_predictions(self, weights, biases, X, Y):
        forward_output_val = forward_propagation_r.forward_propagation_regression(weights, biases, X)
        key, value = list(forward_output_val.items())[-1]
        Y_val_predicted = value.flatten()
        rmse_val = self.root_mean_squared_error(Y_val_predicted, Y)
        return rmse_val


    def mean_squared_error(self, Y_predicted, Y):
        return np.mean((Y_predicted - Y)**2)


    def root_mean_squared_error(self, Y_predicted, Y):
        mse = self.mean_squared_error(Y_predicted, Y)
        rmse = np.sqrt(mse)
        return rmse


import numpy as np

class OneHotEncoding:
    
    """
    This class is used for classification problems to perform One hot Encoding on the label
    """

    def one_hot_encoding(self, Y):

        one_hot_Y = np.zeros((Y.size, Y.max() + 1))
        one_hot_Y[np.arange(Y.size), Y] = 1
        one_hot_Y = one_hot_Y.T

        return one_hot_Y


import matplotlib.pyplot as plt
import plotly.graph_objects as go

class Visualization:


    def plot_loss_curve(self, losses):
        plt.plot(losses)
        plt.title('Loss Curve')
        plt.xlabel('Epochs')
        plt.ylabel('Loss')
        plt.show()



    def plot_accuracy_curve(self, accuracies):
        plt.plot(accuracies)
        plt.title('Accuracy Curve')
        plt.xlabel('Epochs')
        plt.ylabel('Accuracy')
        plt.show()

    

    def plot_neural_network_architecture(self, n_input, n_layers, n_neurons_per_layer, n_output):
        fig = go.Figure()

        fig.add_trace(go.Scatter(x=[0.5] * n_input, y=[i for i in range(n_input)], mode='markers', marker=dict(size=30),
                                text=[f'Feature {i}' for i in range(1, n_input + 1)], name='Input Layer'))

        for i, neurons in enumerate(n_neurons_per_layer, start=1):
            fig.add_trace(go.Scatter(x=[i] * neurons, y=[j for j in range(neurons)],
                                    mode='markers', marker=dict(size=30),
                                    text=[f'Neuron {j + 1}' for j in range(neurons)],
                                    name=f'Hidden Layer {i}'))

        fig.add_trace(go.Scatter(x=[n_layers + 1] * n_output, y=[k for k in range(n_output)],
                                mode='markers', marker=dict(size=30),
                                text=[f'Class {k + 1}' for k in range(n_output)], name='Output Layer'))

        fig.update_layout(title='Neural Network Architecture',
                          xaxis_title='Layers',
                          yaxis_title='Neurons',
                          showlegend=True)

        fig.show()


from setuptools import setup, find_packages

setup(
    name="NeuroInsight",
    author="Pranjal Khadka",
    author_email="pranzalkhadka1@gmail.com",
    version="0.1",
    description="A custom neural network library for educational purposes",
    long_description="NeuroInsight is a simple neural network library designed to help beginners understand the internals of neural networks.",
    packages=find_packages(),
    install_requires=[
        'numpy',
        'matplotlib',
        'plotly'
    ],
)
